# Copyright 2014  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging

import cliapp

import distixlib


class MetadataManipulationPlugin(cliapp.Plugin):

    def enable(self):
        self.app.add_subcommand(
            '_list', self.list, arg_synopsis='FILENAME', hidden=True)
        self.app.add_subcommand(
            '_get', self.get, arg_synopsis='FILENAME KEY...', hidden=True)
        self.app.add_subcommand(
            '_set', self.set, arg_synopsis='FILENAME KEY=VALUE...',
            hidden=True)

    def list(self, args):
        '''Show all key/value pairs in a metadata file.'''

        loader = distixlib.MetadataLoader()
        for filename in args:
            metadata = loader.load_from_file(filename)
            self._list_all_metadata(metadata)

    def get(self, args):
        '''Show all values for the given keys.'''

        filename = args[0]
        loader = distixlib.MetadataLoader()
        metadata = loader.load_from_file(filename)
        for key in args[1:]:
            self._list_values_for_key(metadata, key)

    def set(self, args):
        filename = args[0]

        metadata = self._load_metadata(filename)
        for pair in args[1:]:
            key, value = self._parse_pair(pair)
            metadata.add(key, value)

        self._save_metadata(filename, metadata)

    def _parse_pair(self, pair):
        return pair.split('=', 1)

    def _load_metadata(self, filename):
        logging.debug('loading from %s', filename)
        loader = distixlib.MetadataLoader()
        try:
            metadata = loader.load_from_file(filename)
        except distixlib.CannotLoadMetadataError:
            logging.debug(
                'Could not load metadata from %s, assuming empty',
                filename)
            metadata = distixlib.Metadata()
        return metadata

    def _save_metadata(self, filename, metadata):
        saver = distixlib.MetadataSaver()
        saver.save_to_file(filename, metadata)

    def _list_all_metadata(self, metadata):
        for key in sorted(metadata.keys()):
            self._list_values_for_key(metadata, key)

    def _list_values_for_key(self, metadata, key):
        for value in sorted(metadata.get_all_values(key)):
            self.app.output.write('%s=%s\n' % (key, value))
